# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""API-wide URLs."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("BASE_URL", "CDN_URL", "OAUTH2_API_URL", "REST_API_URL", "VERSION")

import typing

VERSION: typing.Final[int] = 10
"""The Gateway and REST version to use."""

BASE_URL: typing.Final[str] = "https://discord.com"
"""The base URL."""

REST_API_URL: typing.Final[str] = f"{BASE_URL}/api/v{VERSION}"
"""The REST API URL."""

OAUTH2_API_URL: typing.Final[str] = f"{REST_API_URL}/oauth2"
"""The OAUTH2 API URL."""

CDN_URL: typing.Final[str] = "https://cdn.discordapp.com"
"""The CDN URL."""

MEDIA_PROXY_URL: typing.Final[str] = "https://media.discordapp.net"
"""The media proxy URL."""
