# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""A little helpful script to increment the package version after every release."""

from __future__ import annotations

import pathlib
import sys

sys.path.append(str(pathlib.Path.cwd()))

from hikari.internal import ux

version = ux.HikariVersion(sys.argv[1])

if version.prerelease is not None:
    # Increase the pre-release number if there
    prerelease_str, prerelease_num = version.prerelease
    version.prerelease = (prerelease_str, prerelease_num + 1)

else:
    # Or add it if missing and bump patch version
    major, minor, patch = version.version
    version.version = (major, minor, patch + 1)
    version.prerelease = (".dev", 0)

print(version)
